from __future__ import annotations

import argparse
import sys

import matplotlib.pyplot as plt  # type: ignore[import-not-found]
import numpy as np
import xarray as xr

# Define longitude choices.
grids_theta_mlt = (
    "12",
    "",
    "14",
    "",
    "16",
    "",
    "18",
    "",
    "20",
    "",
    "22",
    "",
    "0",
    "",
    "2",
    "",
    "4",
    "",
    "6",
    "",
    "8",
    "",
    "10",
    "",
)
grids_theta_deg = (
    "0",
    "",
    "30",
    "",
    "60",
    "",
    "90",
    "",
    "120",
    "",
    "150",
    "",
    "180",
    "",
    "210",
    "",
    "240",
    "",
    "270",
    "",
    "300",
    "",
    "330",
    "",
)


class InvalidLatitudesException(Exception):
    """Custom exception for invalid latitudes."""


# Exit the program if given invalid latitudes.
def lats_invalid() -> None:
    error_message = (
        "Invalid latitudes! Please enter one of the following:\n"
        "1) '-n' or '-s' as an option, OR\n"
        "2) 0 <= lats_min < lats_max <= 90 for the northern hemisphere, OR\n"
        "3) -90 <= lats_min < lats_max <= 0 for the southern hemisphere."
    )
    sys.exit(error_message)


# Return a tuple of the plot parameters.
def get_plot_params(
    lats_max: int, lats_min: int, spacing: int, da: xr.DataArray
) -> tuple[range, tuple[str, ...], xr.DataArray]:
    if 0 <= lats_min < lats_max <= 90:
        range_r = range(int(90 - lats_max), int(lats_max - lats_min), spacing)
        grids_r = tuple(
            f"{lat}" for lat in range(int(lats_max), int(lats_min), -spacing)
        )
        coord_ns = da.ggcm.coords["colats"]
    elif -90 <= lats_min < lats_max <= 0:
        range_r = range(int(lats_min), int(lats_max), spacing)
        grids_r = tuple(
            f"{lat}" for lat in range(int(lats_min), int(lats_max), spacing)
        )
        coord_ns = da.coords["lats"]
    else:
        error_message = (
            "Invalid latitudes! Please enter one of the following:\n"
            "1) 0 <= lats_min < lats_max <= 90 for the northern hemisphere, OR\n"
            "2) -90 <= lats_min < lats_max <= 0 for the southern hemisphere."
        )
        raise InvalidLatitudesException(error_message)
    return range_r, grids_r, coord_ns


def plot_from_dataarray(
    da: xr.DataArray,
    lats_max: int,
    lats_min: int,
    spacing: int,
    mlt: bool,
    levels=None,
    cmap="bwr",
    extend="both",
    **kwargs,
) -> None:
    da_sliced = da.sel(lats=slice(lats_max, lats_min))
    if levels is None:
        abs_max = np.abs(da_sliced.values).max()
        levels = np.linspace(-abs_max, abs_max, 51)

    plot_title = f"{da.attrs['long_name']} [{da.attrs['units']}]"
    lon = da_sliced.coords["longs"]

    fig, ax = plt.subplots(
        subplot_kw={"projection": "polar", "theta_offset": np.pi / 2}
    )
    range_theta = range(0, 360, 15)
    plt.thetagrids(range_theta, grids_theta_mlt if mlt else grids_theta_deg)

    range_r, grids_r, coord_ns = get_plot_params(lats_max, lats_min, spacing, da_sliced)
    plt.rgrids(range_r, grids_r)
    ax.set_title(plot_title)

    mesh = ax.contourf(
        np.deg2rad(lon),
        coord_ns,
        da_sliced.T,
        cmap=cmap,
        levels=levels,
        extend=extend,
        **kwargs,
    )
    fig.colorbar(mesh)
    plt.show()


def get_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser(
        description=("This module plots data from an OpenGGCM output file.")
    )

    # Define mutually exclusive options.
    group = parser.add_mutually_exclusive_group()
    group.add_argument("-n", "--north", action="store_true")
    group.add_argument("-s", "--south", action="store_true")

    # Define positional arguments.
    parser.add_argument("file", help="data file")
    parser.add_argument("var", help="variable to be plotted")
    parser.add_argument("lats_max", type=int, nargs="?", help="maximum latitude")
    parser.add_argument("lats_min", type=int, nargs="?", help="minimum latitude")
    parser.add_argument(
        "spacing",
        type=int,
        nargs="?",
        default=10,
        help="spacing between grids",
    )
    parser.add_argument(
        "mlt",
        nargs="?",
        default="true",
        help="magnetic local time coordinates on/off",
    )

    args = parser.parse_args()

    # Validate the arguments.
    if (args.north or args.south) and (args.lats_max or args.lats_min):
        lats_invalid()

    # Define conditionally default latitudes.
    if args.south:
        if args.lats_max is None:
            args.lats_max = -50
        if args.lats_min is None:
            args.lats_min = -90
    else:  # Default to northern hemisphere.
        if args.lats_max is None:
            args.lats_max = 90
        if args.lats_min is None:
            args.lats_min = 50

    return args


def main() -> None:
    args = get_args()
    try:
        with xr.open_dataset(args.file) as ds:
            ds[args.var].ggcm.plot(
                lats_max=args.lats_max,
                lats_min=args.lats_min,
                spacing=args.spacing,
                mlt=(args.mlt.lower() == "true"),
            )
    except InvalidLatitudesException:
        lats_invalid()


if __name__ == "__main__":
    main()



# mostly taken from viscid... thanks Kris
from __future__ import annotations

import contextlib
import logging
import os
from collections import OrderedDict
from collections.abc import Mapping
from typing import Any

import numpy as np
from numpy.typing import NDArray

from ggcmpy import _jrrle, openggcm  # type: ignore[attr-defined]

from .fortran_file import FortranFile

logger = logging.getLogger(__name__)

read_ascii = False

_jrrle_read_func = (
    _jrrle.read_jrrle1d,
    _jrrle.read_jrrle2d,
    _jrrle.read_jrrle3d,
)


def _jrrle_read_field(file, fld_name: str, meta: dict[str, Any]) -> NDArray[Any]:
    shape = meta["shape"]
    ndim = len(shape)
    arr = np.empty(shape, dtype="float32", order="F")
    success = _jrrle_read_func[ndim - 1](file.unit, arr, fld_name, read_ascii)
    if not success:
        msg = "read_func failed"
        raise RuntimeError(msg)
    return arr


def _jrrle_inquire_next(
    file: FortranFile,
) -> tuple[str, dict[str, Any]] | None:
    b_varname = np.array(" " * 80, dtype="S80")
    b_tstring = np.array(" " * 80, dtype="S80")
    found_field, ndim, nx, ny, nz, it = _jrrle.inquire_next(
        file.unit, b_varname, b_tstring
    )
    if not found_field:
        return None

    attrs = {
        "shape": (nx, ny, nz)[:ndim],
        "varname": str(np.char.decode(b_varname)).strip(),
        "inttime": it,
        "timestr": str(np.char.decode(b_tstring)).strip(),
        "file_position": file.tell(),
    }
    with contextlib.suppress(ValueError):
        # ignore malformed time string (this is the case for other data, like grid files)
        attrs |= openggcm.parse_timestring(attrs["timestr"])

    return attrs["varname"], attrs


class JrrleFile(FortranFile):
    """Interface for actually opening / reading a jrrle file"""

    def __init__(self, filename: str | os.PathLike[Any], mode: str = "r"):
        assert mode == "r"

        self.fields_seen: OrderedDict[str, Any] = OrderedDict()
        self.seen_all_fields = False
        super().__init__(filename)

    def read_field(self, fld_name) -> tuple[Any, NDArray[Any]]:
        """Read a field"""
        logger.debug("read_field(%s)" % fld_name)
        meta = self._inquire(fld_name)
        return meta, _jrrle_read_field(self, fld_name, meta)

    @property
    def vars(self) -> Mapping[str, Mapping[str, Any]]:
        self._inquire_all_fields()
        return self.fields_seen

    def _inquire_all_fields(self) -> None:
        if self.seen_all_fields:
            return

        logger.debug("_inquire_all_fields() (%s)" % self.filename)
        self.rewind()
        while True:
            rv = self._inquire_next()
            if not rv:
                break
            self.advance_one_line()

    def _inquire(self, fld_name: str) -> Any:
        if fld_name in self.fields_seen:
            meta = self.fields_seen[fld_name]
            # For some mysterious reason, the rewind is necessary -- without it,
            # reading after the seek fails. Mysteriously, a self.tell() also
            # sometimes makes things work.
            self.rewind()
            self.seek(meta["file_position"])
            return meta

        if self.fields_seen:
            # seek to past last previously read field
            last_added = next(reversed(self.fields_seen))
            self.seek(self.fields_seen[last_added]["file_position"])
            self.advance_one_line()

        while True:
            rv = self._inquire_next()
            if not rv:
                break
            found_fld_name, meta = rv
            if found_fld_name == fld_name:
                return meta
            self.advance_one_line()

        msg = f"file '{self.filename}' has no field '{fld_name}'"
        raise KeyError(msg)

    def _inquire_next(self) -> tuple[str, Any] | None:
        """Collect the meta-data from the next field in the file

        Returns:
            tuple (field name, dict of meta data) both
            of which will be None if there are no more Fields

        Note:
            After this operation is done, the file-pointer will be reset
            to the position it was before the inquiry.
        """

        rv = _jrrle_inquire_next(self)
        if not rv:
            # end of file
            self.seen_all_fields = True
            return None

        varname, meta = rv
        if varname in self.fields_seen:
            assert meta == self.fields_seen[varname]
        else:
            self.fields_seen[varname] = meta

        return varname, meta



from __future__ import annotations

import numpy as np
import pytest
import xarray as xr

import ggcmpy
from ggcmpy import openggcm

sample_time_array = np.array([[2010, 1, 1, 13, 0, 0, 100], [2010, 1, 1, 13, 1, 0, 100]])
sample_datetime64 = [
    np.datetime64("2010-01-01T13:00:00.100000000"),
    np.datetime64("2010-01-01T13:01:00.100000000"),
]


def test_to_dt64():
    assert np.all(
        openggcm.timearray_to_dt64(sample_time_array[0]) == sample_datetime64[0]
    )


def test_to_timearray():
    assert np.all(
        openggcm._dt64_to_timearray(sample_datetime64, dtype=np.int32)
        == sample_time_array
    )


@pytest.mark.parametrize(
    ("dims", "data_time_array", "data_datetime64"),
    [
        (("time", "time_array"), sample_time_array, sample_datetime64),
        (("time_array",), sample_time_array[0], sample_datetime64[0]),
    ],
)
def test_AmieTimeArrayCoder(dims, data_time_array, data_datetime64):
    var = xr.Variable(dims, data_time_array)
    var_dt64 = xr.Variable(dims[:-1], data_datetime64)

    coder = openggcm.AmieTimeArrayCoder()
    decoded_var = coder.decode(var, "name")
    assert decoded_var.equals(var)  # type: ignore[no-untyped-call]

    var.attrs["units"] = "time_array"
    decoded_var = coder.decode(var, "name")
    assert decoded_var.equals(var_dt64)  # type: ignore[no-untyped-call]
    assert var.attrs == {"units": "time_array"}  # original var not changed
    assert decoded_var.encoding == {"units": "time_array", "dtype": var.dtype}


@pytest.mark.parametrize(
    ("dims", "data_time_array", "data_datetime64"),
    [
        (("time", "time_array"), sample_time_array, sample_datetime64),  # time array
        (("time_array",), sample_time_array[0], sample_datetime64[0]),  # scalar time
    ],
)
def test_encode_openggcm_variable(dims, data_time_array, data_datetime64):
    var = xr.Variable(dims, data_time_array)
    var_dt64 = xr.Variable(dims[:-1], data_datetime64)

    var.attrs["units"] = "time_array"
    coder = openggcm.AmieTimeArrayCoder()
    decoded_var = coder.decode(var, "name")
    assert decoded_var.attrs == {}
    assert decoded_var.equals(var_dt64)  # type: ignore[no-untyped-call]
    encoded_var = coder.encode(decoded_var)
    assert encoded_var.equals(var)  # type: ignore[no-untyped-call]
    assert encoded_var.encoding == {}


def test_coords():
    ds = xr.Dataset(
        {"longs": np.linspace(-180, 180, 61)}, {"lats": np.linspace(90, -90, 181)}
    )
    assert np.allclose(ds.ggcm.coords["mlts"], np.linspace(0, 24, 61))
    assert np.allclose(ds.ggcm.coords["colats"], np.linspace(0, 180, 181))
    assert np.allclose(ds.ggcm.mlts, np.linspace(0, 24, 61))
    assert np.allclose(ds.ggcm.colats, np.linspace(0, 180, 181))


def test_iopar():
    iof = xr.open_dataset(f"{ggcmpy.sample_dir}/coupling0001.iof.000030")
    iof = iof.isel(time=0)
    iox = ggcmpy.iono.iopar(iof)  # type: ignore[no-untyped-call]
    xr.testing.assert_allclose(iox["delbt"], iof["delbt"])


def test_epoch1966():
    testdate = np.datetime64("1967-02-03T04:05:06.100")
    dsecs = ggcmpy.openggcm.epoch1966(testdate)

    import datetime

    date0 = np.datetime64("1966-01-01T00:00:00.000").astype(datetime.datetime)
    assert np.isclose(dsecs, (testdate - date0).total_seconds())


def test_cotr_gse_to_mhd():
    testdate = np.datetime64("1967-02-03T04:05:06.100")

    r1 = np.array([1.0, 2.0, 3.0])
    r2 = ggcmpy.openggcm.cotr(testdate, "gse", "mhd", r1)
    assert np.allclose(r2, [-1.0, -2.0, 3.0])


def test_cotr_gse_to_gsm():
    # special time indicating no dipole tilt
    testdate = np.datetime64("1967-01-01T00:00:00.000")

    r1 = np.array([1.0, 2.0, 3.0])
    r2 = ggcmpy.openggcm.cotr(testdate, "gse", "gsm", r1)
    assert np.allclose(r2, [1.0, 2.0, 3.0])



"""
iono.py

Ionosphere diagnostics
"""

from __future__ import annotations

import xarray as xr

from ggcmpy import _jrrle  # type: ignore[attr-defined]


def gradpt(pot):
    """Compute the gradient of potential field on a spherical surface
    (at ionospheric height) using the Fortran routine.

    Parameters
    ----------
    pot : xarray.DataArray
        2D array of the potential field on a spherical surface.
        Coordinates must be 'longs' and 'lats'.

    Returns
    -------
    xarray.Dataset
        Dataset containing the two components of the gradient field:
        'et' (theta-component) and 'ep' (phi-component).
    """
    pot = pot.transpose("longs", "lats")  # ensure correct order for Fortran

    et, ep = _jrrle.ground_perturbation_m.gradpt(pot.values)
    return xr.Dataset(
        {"et": (("longs", "lats"), et), "ep": (("longs", "lats"), ep)},
        coords=pot.coords,
    )


def iopar(ds):
    """Postprocess ionospheric quantities using the Fortran routine.

    Parameters
    ----------
    ds : xarray.Dataset
        Dataset containing the necessary input variables:
        - fac_tot: FAC
        - pot: Potential
        - sigp: Pedersen conductivity
        - sigh: Hall conductivity

    Returns
    -------
    xarray.Dataset
        Dataset containing the computed ionospheric parameters:
        - ep: Electric field phi-component
        - et: Electric field theta-component
        - ctiot: Total ionospheric current theta-component
        - ctiop: Total ionospheric current phi-component
        - tau:
        - ctaup: current phi-component
        - ctaut: current theta-component
        - cpolp: current phi-component
        - cpolt: current theta-component
        - delbr: ground magnetic perturbation radial component
        - delbp: ground magnetic perturbation phi-component
        - delbt: ground magnetic perturbation theta-component
        - xjh: Joule heating rate"""
    if "time" in ds.sizes:
        msg = "Dataset must be 2D without time dimension. Use .isel(time=0) to select a time slice."
        raise ValueError(msg)
    ds = ds.transpose("longs", "lats")  # ensure correct order for Fortran

    ep, et, ctiot, ctiop, tau, ctaup, ctaut, cpolp, cpolt, delbr, delbp, delbt, xjh = (
        _jrrle.ground_perturbation_m.iopar(ds.fac_tot, ds.pot, ds.sigp, ds.sigh)
    )
    fields = {
        "ep": ep,
        "et": et,
        "ctiot": ctiot,
        "ctiop": ctiop,
        "tau": tau,
        "ctaup": ctaup,
        "ctaut": ctaut,
        "cpolp": cpolp,
        "cpolt": cpolt,
        "delbr": delbr,
        "delbp": delbp,
        "delbt": delbt,
        "xjh": xjh,
    }
    return xr.Dataset(
        {name: (("longs", "lats"), data) for name, data in fields.items()},
        coords=ds.coords,
    )



from __future__ import annotations

import importlib.metadata

project = "ggcmpy"
copyright = "2025, Kai Germaschewski"
author = "Kai Germaschewski"
version = release = importlib.metadata.version("ggcmpy")

extensions = [
    "myst_parser",
    "nbsphinx",
    "sphinx.ext.autodoc",
    "sphinx.ext.intersphinx",
    "sphinx.ext.mathjax",
    "sphinx.ext.napoleon",
    "sphinx.ext.todo",
    "sphinx_autodoc_typehints",
    "sphinx_copybutton",
]

source_suffix = [".rst", ".md"]
exclude_patterns = [
    "_build",
    "**.ipynb_checkpoints",
    "Thumbs.db",
    ".DS_Store",
    ".env",
    ".venv",
]

todo_include_todos = True

# html_theme = "furo"
html_theme = "sphinx_book_theme"
html_title = ""

html_context = {
    "github_user": "unh-hpc",
    "github_repo": "ggcmpy",
    "github_version": "main",
    "doc_path": "docs",
}

html_logo = "openggcm-logo.png"

html_theme_options = dict(  # noqa: C408
    # analytics_id=''  this is configured in rtfd.io
    # canonical_url="",
    repository_url="https://github.com/unh-hpc/ggcmpy",
    repository_branch="main",
    navigation_with_keys=False,  # pydata/pydata-sphinx-theme#1492
    navigation_depth=4,
    path_to_docs="docs",
    use_edit_page_button=True,
    use_repository_button=True,
    use_issues_button=True,
    home_page_in_toc=False,
    # extra_footer="""<p>Xarray is a fiscally sponsored project of <a href="https://numfocus.org">NumFOCUS</a>,
    # # a nonprofit dedicated to supporting the open-source scientific computing community.<br>
    # # Theme by the <a href="https://ebp.jupyterbook.org">Executable Book Project</a></p>""",
    # twitter_url="https://twitter.com/xarray_dev",
    # icon_links=[],  # workaround for pydata/pydata-sphinx-theme#1220
    # announcement="<a href='https://forms.gle/KEq7WviCdz9xTaJX6'>Xarray's 2024 User Survey is live now. Please take ~5 minutes to fill it out and help us improve Xarray.</a>",
)


# html_theme_options: dict[str, Any] = {
#     "footer_icons": [
#         {
#             "name": "GitHub",
#             "url": "https://github.com/unh-hpc/ggcmpy",
#             "html": """
#                 <svg stroke="currentColor" fill="currentColor" stroke-width="0" viewBox="0 0 16 16">
#                     <path fill-rule="evenodd" d="M8 0C3.58 0 0 3.58 0 8c0 3.54 2.29 6.53 5.47 7.59.4.07.55-.17.55-.38 0-.19-.01-.82-.01-1.49-2.01.37-2.53-.49-2.69-.94-.09-.23-.48-.94-.82-1.13-.28-.15-.68-.52-.01-.53.63-.01 1.08.58 1.23.82.72 1.21 1.87.87 2.33.66.07-.52.28-.87.51-1.07-1.78-.2-3.64-.89-3.64-3.95 0-.87.31-1.59.82-2.15-.08-.2-.36-1.02.08-2.12 0 0 .67-.21 2.2.82.64-.18 1.32-.27 2-.27.68 0 1.36.09 2 .27 1.53-1.04 2.2-.82 2.2-.82.44 1.1.16 1.92.08 2.12.51.56.82 1.27.82 2.15 0 3.07-1.87 3.75-3.65 3.95.29.25.54.73.54 1.48 0 1.07-.01 1.93-.01 2.2 0 .21.15.46.55.38A8.013 8.013 0 0 0 16 8c0-4.42-3.58-8-8-8z"></path>
#                 </svg>
#             """,
#             "class": "",
#         },
#     ],
#     "source_repository": "https://github.com/unh-hpc/ggcmpy",
#     "source_branch": "main",
#     "source_directory": "docs/",
# }

myst_enable_extensions = [
    "colon_fence",
]

intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
}

nitpick_ignore = [
    ("py:class", "_io.StringIO"),
    ("py:class", "_io.BytesIO"),
]

always_document_param_types = True

nbsphinx_timeout = 600
nbsphinx_execute = "always"
nbsphinx_allow_errors = False



from __future__ import annotations

import numpy as np

sample_iof = {
    "coords": {"lats", "longs", "time"},
    "data_vars": [
        "prec_e_fe_1",
        "prec_e_e0_1",
        "prec_e_fe_2",
        "prec_e_e0_2",
        "pacurr",
        "fac_tot",
        "pot",
        "delphi",
        "ppio",
        "rrio",
        "ttio",
        "vdown",
        "sigh",
        "sigp",
        "fac_dyn",
        "ctiot",
        "ctiop",
        "tau",
        "ctaup",
        "ctaut",
        "cpolp",
        "cpolt",
        "delbr",
        "delbp",
        "delbt",
        "epio",
        "etio",
        "xjh",
        "elapsed_time",
        "inttime",
    ],
    "time": np.asarray(
        ["1967-01-01T00:00:30.150", "1967-01-01T00:01:00.101"], dtype=np.datetime64
    ),
    "sizes": {"lats": 181, "longs": 61},
}

sample_3df = {
    "coords": {"x", "y", "z", "time"},
    "data_vars": [
        "rr",
        "pp",
        "vx",
        "vy",
        "vz",
        "bx",
        "by",
        "bz",
        "xjx",
        "xjy",
        "xjz",
        "xtra1",
        "xtra2",
        "resis",
        "elapsed_time",
        "inttime",
    ],
    "time": np.asarray(["1967-01-01T00:20:00.033000"], dtype=np.datetime64),
    "sizes": {"x": 64, "y": 32, "z": 32},
}

sample_py = {
    "coords": {"x", "y", "z", "time"},
    "data_vars": [
        "rr",
        "pp",
        "vx",
        "vy",
        "vz",
        "bx",
        "by",
        "bz",
        "xjx",
        "xjy",
        "xjz",
        "xtra1",
        "xtra2",
        "resis",
        "elapsed_time",
        "inttime",
    ],
    "time": np.asarray(["1967-01-01T00:20:00.033000"], dtype=np.datetime64),
    "sizes": {"x": 64, "z": 32},
}



from __future__ import annotations

import pathlib

import numpy as np
import pytest
from conftest import sample_iof

import ggcmpy
from ggcmpy.backends.jrrle import JrrleFile


def test_jrrle_file_non_existent():
    with pytest.raises(RuntimeError):
        JrrleFile("/non-existent")


def test_jrrle_file_open():
    JrrleFile(str(pathlib.Path(ggcmpy.sample_dir) / "coupling0001.iof.000030"))


def test_jrrle_file_open_path():
    JrrleFile(pathlib.Path(ggcmpy.sample_dir) / "coupling0001.iof.000030")


def test_jrrle_file_iter():
    with JrrleFile(pathlib.Path(ggcmpy.sample_dir) / "coupling0001.iof.000030") as file:
        for var_name, ref_name in zip(file.vars, sample_iof["data_vars"], strict=False):
            assert var_name == ref_name


def test_jrrle_file_inquire():
    vars: list[str] = sample_iof["data_vars"]  # type: ignore[assignment]
    with JrrleFile(pathlib.Path(ggcmpy.sample_dir) / "coupling0001.iof.000030") as file:
        for var_name in vars[:3]:
            file._inquire(var_name)
        assert vars[1] in file.fields_seen
        assert vars[3] not in file.fields_seen
        file._inquire(vars[3])
        assert vars[3] in file.fields_seen


def test_jrrle_file_read_field():
    vars: list[str] = sample_iof["data_vars"]  # type: ignore[assignment]
    with JrrleFile(pathlib.Path(ggcmpy.sample_dir) / "coupling0001.iof.000030") as file:
        _, var0 = file.read_field(vars[0])
        assert np.isclose(np.sum(var0), 0.014970759)
        _, var1 = file.read_field(vars[1])
        assert np.isclose(np.sum(var1), 36606.758)
        _, var2 = file.read_field(vars[2])
        assert np.isclose(np.sum(var2), 7.3803894e-05)

        _, var1 = file.read_field(vars[1])
        assert np.isclose(np.sum(var1), 36606.758)

        _, var3 = file.read_field(vars[3])
        assert np.isclose(np.sum(var3), 38640.805)

        with pytest.raises(KeyError):
            file.read_field("nowhere")


def test_jrrle_read_grid():
    with JrrleFile(
        pathlib.Path(ggcmpy.sample_dir) / "djl06008" / "djl06008.grid"
    ) as file:
        assert {"gridx", "gridy", "gridz"} <= file.vars.keys()



from __future__ import annotations

from pathlib import Path

from . import iono
from ._version import version as __version__

sample_dir = Path(__file__).parent / "sample"

__all__ = [
    "__version__",
    "iono",
    "sample_dir",
]



from __future__ import annotations

import pathlib
from warnings import warn

import pandas as pd

SATELLITES = {
    "om1997": "OMNI 1997",
    "wi": "Wind",
}

QUANTITIES: dict[str, dict[str, str | float]] = {
    "vxgse": {
        "name": "Vx (GSE)",
        "long_name": "Vx in GSE Coordinates",
        "units": "km/s",
    },
    "vygse": {
        "name": "Vy (GSE)",
        "long_name": "Vy in GSE Coordinates",
        "units": "km/s",
    },
    "vzgse": {
        "name": "Vz (GSE)",
        "long_name": "Vz in GSE Coordinates",
        "units": "km/s",
    },
    "bxgse": {"name": "Bx (GSE)", "long_name": "Bx in GSE Coordinates", "units": "nT"},
    "bygse": {"name": "By (GSE)", "long_name": "By in GSE Coordinates", "units": "nT"},
    "bzgse": {"name": "Bz (GSE)", "long_name": "Bz in GSE Coordinates", "units": "nT"},
    "rr": {"name": "N", "long_name": "Number Density", "units": "cm^-3"},
    "pp": {"name": "P", "long_name": "Pressure", "units": "nPa"},
    "tt": {"name": "T", "long_name": "Temperature", "factor": 1e-3, "units": "keV"},
}


def read_ggcm_solarwind_file(
    filename: pathlib.Path, varname: str | None = None
) -> pd.DataFrame:
    """Reads a single field time series (e.g., "wi.bzgse") file into a pandas DataFrame.

    Parameters
    ----------
    filename : pathlib.Path
        Path to the time series file.
    varname : str | None, optional
        Name of the variable. If None, the filename is used.

    Returns
    -------
    pd.DataFrame
        A pandas DataFrame containing the time series data.
    """
    if varname is None:
        varname = filename.name
    data = pd.read_csv(
        filename,
        sep=r"\s+",
        names=["year", "month", "day", "hour", "minute", "second", varname],
    )
    if data["year"].max() < 100:
        data["year"] += 1900
    data["datetime"] = pd.to_datetime(
        data[["year", "month", "day", "hour", "minute", "second"]]
    )
    data = data.set_index("datetime")

    data = data[[varname]]  # keep only the variable column
    if data.index.has_duplicates:
        warn(
            f"duplicate times found in {filename}, dropping duplicates",
            stacklevel=2,
        )
        # Remove duplicate times, keeping the first occurrence
        data = data[~data.index.duplicated(keep="first")]
    sat, sfx = varname.split(".")  # e.g., om1997.vxgse
    sat = SATELLITES.get(sat, sat)
    if sfx in QUANTITIES:
        quantity = QUANTITIES[sfx]
        if "factor" in quantity:
            data[varname] *= quantity["factor"]
        data[varname].attrs["long_name"] = f"{sat} {quantity['long_name']}"
        data[varname].attrs["name"] = quantity["name"]
        data[varname].attrs["units"] = quantity["units"]

    return data


def read_ggcm_solarwind_directory(directory: pathlib.Path, glob: str = "*"):
    """Reads all single field time series files in a GGCM solar wind directory into a pandas DataFrame.

    Parameters
    ----------
    directory : pathlib.Path
        Path to the directory containing time series files.
    glob : str, optional
        Glob pattern to match files, by default "*".

    Returns
    -------
    pd.DataFrame
        A pandas DataFrame containing the concatenated time series data from all files.
    """

    dfs, attrs = [], {}
    for file in directory.glob(glob):
        df_read = read_ggcm_solarwind_file(file)
        varname = df_read.columns[0]
        dfs.append(df_read)
        attrs[varname] = df_read[varname].attrs

    df_combined = pd.concat(dfs, axis=1)
    for varname in df_combined.columns:
        df_combined[varname].attrs = attrs[varname]

    return df_combined


def store_to_pyspedas(df: pd.DataFrame):
    """Stores a pandas DataFrame into pyspedas tplot variable.

    Parameters
    ----------
    df : pd.DataFrame
        DataFrame to be stored.
    """

    try:
        # pylint: disable=C0415
        import pyspedas  # type: ignore[import-not-found]
    except ImportError as err:
        msg = "pyspedas is required for this function. Please install it first."
        raise ImportError(msg) from err

    for varname in df.columns:
        pyspedas.store_data(varname, data={"x": df.index, "y": df[varname]})
        da = pyspedas.get_data(varname, xarray=True)
        attrs = df[varname].attrs
        if "long_name" in attrs:
            da.attrs["plot_options"]["yaxis_opt"]["axis_label"] = attrs["name"]
        if "name" in attrs:
            da.attrs["plot_options"]["yaxis_opt"]["legend_names"] = [attrs["name"]]
        if "units" in attrs:
            da.attrs["plot_options"]["yaxis_opt"]["axis_subtitle"] = attrs["units"]
            # pyspedas.set_units(varname, attrs["units"]) # FIXME, doesn't seem to work

        if varname.endswith(".pp"):
            pyspedas.options(varname, "ylog", True)



from __future__ import annotations

import logging
from collections.abc import Mapping
from typing import TYPE_CHECKING, Any

import numpy as np
from numpy.typing import NDArray
from xarray.backends.common import BackendArray
from xarray.core import indexing

if TYPE_CHECKING:
    from .jrrle_store import JrrleStore

logger = logging.getLogger(__name__)


class JrrleArray(BackendArray):
    """Lazy evaluation of a variable stored in an adios2 file.

    This also takes care of slicing out the specific component of the data stored as 4-d array.
    """

    def __init__(
        self, variable_name: str, datastore: JrrleStore, fld_info: Mapping[str, Any]
    ) -> None:
        self.variable_name = variable_name
        self.datastore = datastore
        self.shape = fld_info["shape"]
        self.dtype = np.dtype(np.float32)

    def get_array(self, needs_lock: bool = True) -> NDArray[Any]:
        _, arr = self.datastore.acquire(needs_lock).read_field(self.variable_name)
        return arr

    def __getitem__(self, key: indexing.ExplicitIndexer) -> NDArray[Any]:
        return indexing.explicit_indexing_adapter(  # type: ignore[no-any-return]
            key, self.shape, indexing.IndexingSupport.BASIC, self._getitem
        )

    def _getitem(self, key) -> NDArray[Any]:
        with self.datastore.lock:
            return self.get_array(needs_lock=False)[key]  # type: ignore[no-any-return]



from __future__ import annotations

import logging

import adios2py
import numpy as np
import pytest
import xarray as xr
from conftest import sample_3df, sample_iof, sample_py
from xarray_adios2 import Adios2Store

import ggcmpy
from ggcmpy import openggcm

logger = logging.getLogger(__name__)


def test_open_dataset_iof():
    ds = xr.open_dataset(f"{ggcmpy.sample_dir}/coupling0001.iof.000030")
    assert set(ds.coords.keys()) >= set(sample_iof["coords"])
    assert set(ds.keys()) == set(sample_iof["data_vars"])
    assert ds.time == sample_iof["time"][0]  # type: ignore[index]
    assert ds.pot.sizes == sample_iof["sizes"]


def test_open_dataset_3df():
    ds = xr.open_dataset(f"{ggcmpy.sample_dir}/sample_jrrle.3df.001200")
    assert set(ds.coords.keys()) >= set(sample_3df["coords"])
    assert set(ds.keys()) == set(sample_3df["data_vars"])
    assert ds.time == sample_3df["time"]
    assert ds.rr.sizes == sample_3df["sizes"]


def test_open_dataset_py():
    ds = xr.open_dataset(f"{ggcmpy.sample_dir}/sample_jrrle.py_0.001200")
    assert set(ds.coords.keys()) >= set(sample_py["coords"])
    assert set(ds.keys()) == set(sample_py["data_vars"])
    assert ds.time == sample_py["time"]
    assert ds.rr.sizes == sample_py["sizes"]


def test_read_iof_jrrle_mfdataset():
    filenames = [
        f"{ggcmpy.sample_dir}/coupling0001.iof.{step:06d}" for step in [30, 60]
    ]
    logger.debug("before open")
    ds = xr.open_mfdataset(filenames)
    logger.debug("after open")
    assert set(ds.coords.keys()) >= set(sample_iof["coords"])
    assert set(ds.keys()) == set(sample_iof["data_vars"])
    assert np.all(ds.time == sample_iof["time"])
    logger.debug("before var0")
    var0 = ds[sample_iof["data_vars"][0]]  # type: ignore[index]
    var0 = var0.isel(time=0)
    logger.debug("before to_numpy()")
    assert np.isclose(np.sum(var0), 0.014970759)


@pytest.fixture
def time_dataset():
    amie_encoding = {"units": "time_array", "dtype": "int32"}

    return xr.Dataset(
        {
            "time_scalar": xr.Variable(
                (), np.datetime64("2021-01-01", "ns"), encoding=amie_encoding
            )
        },
        coords={
            "time": xr.Variable(
                "time",
                [np.datetime64("2020-01-01", "ns"), np.datetime64("2020-01-02")],
                encoding=amie_encoding,
            )
        },
    )


def test_time_write_read_by_step(tmp_path, time_dataset):
    filename = tmp_path / "test_ta.bp"
    time_dataset.attrs["step_dimension"] = "time"
    time_dataset.dump_to_store(
        Adios2Store.open(filename, "w"), encoder=openggcm.encode_openggcm
    )
    ds = xr.open_dataset(filename, decode_times=openggcm.AmieTimeArrayCoder())
    # FIXME, should mark time_scalar as invariant, so it doesn't get read back twice
    ds["time_scalar"] = ds.time_scalar.isel(time=0)
    assert time_dataset.equals(ds)


def test_time_write_read_one_step(tmp_path, time_dataset):
    filename = tmp_path / "test_ta_one.bp"
    time_dataset.dump_to_store(
        Adios2Store.open(filename, "w"), encoder=openggcm.encode_openggcm
    )
    with adios2py.File(filename, "r") as file:
        for step in file.steps:
            ds_read = xr.open_dataset(
                Adios2Store(step), decode_times=openggcm.AmieTimeArrayCoder()
            )
            assert time_dataset.equals(ds_read)


def test_fortran_closed():
    """Reproduce issue with FortranFile being closed"""
    ds_3d = xr.open_dataset(ggcmpy.sample_dir / "sample_jrrle.3df.001200")
    ds_3d.pp.sel(y=slice(-20, 20, 5)).plot(x="x", y="z", col="y")
    ds_3d = xr.open_dataset(ggcmpy.sample_dir / "sample_jrrle.3df.001200")
    ds_3d.pp.sel(y=slice(-20, 20, 5)).plot(x="x", y="z", col="y")
    ds_3d = xr.open_dataset(ggcmpy.sample_dir / "sample_jrrle.3df.001200")
    ds_3d.pp.sel(y=slice(-20, 20, 5)).plot(x="x", y="z", col="y")


def test_read_grid():
    ds = xr.open_dataset(f"{ggcmpy.sample_dir}/djl06008/djl06008.py_0.020880")
    assert ds.x.size == 365
    assert ds.z.size == 120



# mostly taken from viscid... thanks Kris and Matt
# mostly stolen from pyggcm... thanks Matt
from __future__ import annotations

import collections
import os
from threading import Lock

from typing_extensions import Any, Self

from ggcmpy import _jrrle  # type: ignore[attr-defined]

# this lock is to prevent multiple threads from grabbing the same
# fortran file unit since checking for available units and opening
# the file may not be atomic. this should not be an issue for multiple
# processes
fortfile_open_lock = Lock()

_fortran_units = collections.defaultdict[int, int](int)


class FortranFile:
    """
    small wrapper to allow the manipulation of fortran files from python
    """

    _unit: None | int = None

    def __init__(self, name: str | os.PathLike[Any], debug: int = 0):
        self.filename = os.fspath(name)
        self.debug = debug
        self._open()

    # Make sure we close it when we're done
    def __del__(self) -> None:
        self.close()

    def _open(self) -> None:
        if self.isopen:
            msg = f"Fortran file '{self.filename}' already open"
            raise RuntimeError(msg)

        with fortfile_open_lock:
            unit: int = _jrrle.fopen(self.filename, debug=self.debug)
            if unit < 0:
                msg = f"Fortran open error ({unit}) on '{self.filename}'"
                raise RuntimeError(msg)
            self._unit = unit
            _fortran_units[self._unit] += 1

    def close(self) -> None:
        if self.isopen:
            assert self._unit is not None
            _fortran_units[self._unit] -= 1
            if _fortran_units[self._unit] == 0:
                _jrrle.fclose(self._unit, debug=self.debug)
                del _fortran_units[self._unit]
            self._unit = None

    def seek(self, offset: int, whence: int = 0) -> int:
        status = _jrrle.seek(self.unit, offset, whence)
        if status != 0:
            msg = f"status != 0: {status}"
            raise AssertionError(msg)
        return status  # type: ignore[no-any-return]

    def tell(self) -> int:
        pos = _jrrle.tell(self.unit)
        assert pos >= 0
        return pos  # type: ignore[no-any-return]

    @property
    def isopen(self) -> bool:
        if self._unit is None:
            return False

        if bool(_jrrle.fisopen(self._unit)):
            return True

        msg = "File has a valid unit, but fortran says it's closed?"
        raise RuntimeError(msg)

    @property
    def unit(self) -> int:
        assert self.isopen
        assert self._unit is not None  # implied by the above, but again for typing
        return self._unit

    def rewind(self) -> None:
        _jrrle.frewind(self.unit, debug=self.debug)

    def advance_one_line(self) -> int:
        return _jrrle.fadvance_one_line(self.unit, debug=self.debug)  # type: ignore[no-any-return]

    def backspace(self) -> None:
        _jrrle.fbackspace(self.unit, debug=self.debug)

    def __enter__(self) -> Self:
        return self

    def __exit__(self, exc_type: Any, exc_value: Any, traceback: Any) -> None:
        self.close()



from __future__ import annotations

import datetime as dt
import os
from collections.abc import Hashable, Mapping
from itertools import islice
from pathlib import Path
from typing import Any
from warnings import warn

import numpy as np
import pandas as pd
import xarray as xr
from numpy.typing import ArrayLike, DTypeLike, NDArray
from typing_extensions import override
from xarray.coding.times import CFDatetimeCoder

from ggcmpy.backends import jrrle

from . import (  # type: ignore[attr-defined]
    _jrrle,
    plot_polar,
)


def read_grid(filename: os.PathLike[Any] | str) -> dict[str, NDArray[Any]]:
    with jrrle.JrrleFile(filename) as file:
        return {
            "x": file.read_field("gridx")[1],
            "y": file.read_field("gridy")[1],
            "z": file.read_field("gridz")[1],
        }


def read_grid2(filename: os.PathLike[Any] | str) -> dict[str, NDArray[Any]]:
    # load the cell centered grid
    with Path(filename).open(encoding="utf-8") as fin:
        nx = int(next(fin).split()[0])
        gx = list(islice(fin, 0, nx, 1))

        ny = int(next(fin).split()[0])
        gy = list(islice(fin, 0, ny, 1))

        nz = int(next(fin).split()[0])
        gz = list(islice(fin, 0, nz, 1))

    return {
        "x": np.array(gx, dtype="f4"),
        "y": np.array(gy, dtype="f4"),
        "z": np.array(gz, dtype="f4"),
    }


def parse_timestring(timestr: str) -> dict[str, Any]:
    if not timestr.startswith("time="):
        msg = f"Time string '{timestr}' is malformed"
        raise ValueError(msg)
    time_comps = timestr.removeprefix("time=").split()

    return {
        "elapsed_time": float(time_comps[0]),
        "time": np.datetime64(
            dt.datetime.strptime(time_comps[2], "%Y:%m:%d:%H:%M:%S.%f")
        ),
    }


def decode_openggcm(ds: xr.Dataset) -> xr.Dataset:
    warn(
        "decode_openggcm() is deprecated. Pass 'decode_times=openggcm.AmieTimeArrayCoder()' "
        "to open_dataset() instead.",
        DeprecationWarning,
        stacklevel=2,
    )
    coder = AmieTimeArrayCoder()
    for name, var in ds.variables.items():
        ds[name] = coder.decode(var, name)

    return ds


def encode_openggcm(
    variables: Mapping[str, xr.Variable], attributes: Mapping[str, Any]
) -> tuple[Mapping[str, xr.Variable], Mapping[str, Any]]:
    coder = AmieTimeArrayCoder()
    new_variables = {name: coder.encode(var, name) for name, var in variables.items()}

    return new_variables, attributes


class AmieTimeArrayCoder(CFDatetimeCoder):
    """
    A custom coder for encoding and decoding time arrays in xarray Variables.
    This class extends the CFDatetimeCoder to handle variables with a custom
    "time_array" encoding, and otherwise falls back to the usual CF coding/encoding.

    Methods
    -------
    encode(variable: xr.Variable, name: Hashable | None = None) -> xr.Variable
        Encodes an xarray Variable with a "time_array" unit into a custom format.
    decode(variable: xr.Variable, name: Hashable | None = None) -> xr.Variable
        Decodes an xarray Variable with a "time_array" unit from a custom format.
    """

    def __init__(self, use_cftime: bool | None = None):
        super().__init__(use_cftime=use_cftime)

    @override
    def encode(
        self, variable: xr.Variable, name: Hashable | None = None
    ) -> xr.Variable:
        if variable.encoding.get("units") == "time_array":
            attrs = variable.attrs.copy()
            attrs["units"] = "time_array"
            attrs.pop("_FillValue", None)

            return xr.Variable(
                dims=(*variable.dims, "time_array"),
                data=_dt64_to_timearray(
                    variable,
                    variable.encoding.get("dtype", "int32"),
                ),
                attrs=attrs,
            )

        return super().encode(variable, name)

    @override
    def decode(
        self, variable: xr.Variable, name: Hashable | None = None
    ) -> xr.Variable:
        if variable.attrs.get("units") == "time_array":
            new_var: xr.Variable = xr.apply_ufunc(
                timearray_to_dt64,
                variable,
                input_core_dims=[["time_array"]],
                vectorize=True,
            )

            new_var.attrs = variable.attrs.copy()
            new_var.encoding = {
                "units": new_var.attrs.pop("units"),
                "dtype": variable.dtype,
            }
            return new_var

        return super().decode(variable, name)


def timearray_to_dt64(time: NDArray[Any]) -> np.datetime64:
    return np.datetime64(
        dt.datetime(
            year=time[0],
            month=time[1],
            day=time[2],
            hour=time[3],
            minute=time[4],
            second=time[5],
            microsecond=time[6] * 1000,
        ),
        "ns",
    )


def _dt64_to_timearray(times: ArrayLike, dtype: DTypeLike) -> ArrayLike:
    dt_times = pd.to_datetime(np.asarray(times))
    return np.array(  # type: ignore[no-any-return]
        [
            dt_times.year,
            dt_times.month,
            dt_times.day,
            dt_times.hour,
            dt_times.minute,
            dt_times.second,
            dt_times.microsecond // 1000,
        ],
        dtype=dtype,
    ).T


@xr.register_dataarray_accessor("ggcm")  # type: ignore[no-untyped-call]
@xr.register_dataset_accessor("ggcm")  # type: ignore[no-untyped-call]
class OpenGGCMAccessor:
    """
    Xarray accessor to add OpenGGCM-specific features

    As of now, that is ds.ggcm.coords which provides mlts and colats.
    """

    def __init__(self, xarray_obj: Any):
        self._obj = xarray_obj
        self._coords: xr.Coordinates = self._obj.coords

        if "colats" not in self._coords and "lats" in self._coords:
            self._coords = self._coords.assign(colats=90 - self._coords["lats"])

        if "mlts" not in self._coords and "longs" in self._coords:
            self._coords = self._coords.assign(
                mlts=(self._coords["longs"] + 180) * 24 / 360,
            )

    def __repr__(self) -> str:
        return "OpenGGCM accessor\n" + repr(self._coords)

    def plot(
        self,
        lats_max: int = 90,
        lats_min: int = 50,
        spacing: int = 10,
        mlt: bool = True,
        **kwargs: Any,
    ) -> None:
        if isinstance(self._obj, xr.Dataset):
            error_message = "Call .ggcm.plot() on a DataArray, not a Dataset."
            raise TypeError(error_message)
        plot_polar.plot_from_dataarray(
            self._obj,
            lats_max=lats_max,
            lats_min=lats_min,
            spacing=spacing,
            mlt=mlt,
            **kwargs,
        )

    @property
    def coords(self) -> xr.Coordinates:
        return self._coords

    def __getattr__(self, name: str) -> Any:
        if name in self._coords:
            return self._coords[name]

        msg = f"Attribute {name} not found"
        raise AttributeError(msg)


def epoch1966(dt64: np.datetime64) -> float:
    """Convert a numpy datetime64 to seconds since 1966-01-01T00:00:00.

    This is the time representation  used in OpenGGCM, and is implemented here
    via the Fortran routine epoch1966() in the f2py module ggcmpy._jrrle.f2py.

    Parameters
    ----------
    dt : np.datetime64
        The datetime to convert.

    Returns
    -------
    float
        The number of seconds since 1966-01-01T00:00:00.
    """
    date = pd.to_datetime(dt64)
    dsecs: float = _jrrle.f2py.epoch1966(
        date.year,
        date.month,
        date.day,
        date.hour,
        date.minute,
        date.second + date.microsecond / 1e6,
    )
    return dsecs


def cotr(date: np.datetime64, cfr: str, cto: str, r1: ArrayLike) -> NDArray[Any]:
    """Transform coordinates between different coordinate systems.

    This function wraps the Fortran routine cotr() in the f2py module ggcmpy._jrrle.f2py.
    takes vector r1 in coordinate system 'cfr' and returns the corresponding
    vector in coordinate system 'cto'

    Coordinate systems 'cfr' and 'cto' can be any of the following:
    'gei'  :  geocentric equatorial inertial
    'geo'  :  geographic
    'gse'  :  geocentric solar ecliptic
    'gsm'  :  geocentric solar magnetospheric
    'sm '  :  solar magnetic
    'mag'  :  geomagnetic
    'mhd'  :  global mhd simulation system, like 'gse' with
              x and y axes mirrored

    Parameters
    ----------
    date : np.datetime64
        The datetime at which to perform the transformation.
    cfr : str
        The coordinate system of the input coordinates.
    cto : str
        The coordinate system of the output coordinates.
    r1 : ArrayLike
        The input coordinates to transform. This should be a 1D array-like of length 3.

    Returns
    -------
    NDArray[Any]
        The transformed coordinates as a numpy array of shape (3,).
    """
    dsecs = epoch1966(date)
    res = _jrrle.f2py.cotr(dsecs, cfr, cto, r1)
    assert isinstance(res, np.ndarray)
    return res



from __future__ import annotations

import numpy as np
import pandas as pd
import pytest

import ggcmpy
from ggcmpy.timeseries import read_ggcm_solarwind_file


def test_read_ggcm_solarwind_file():
    rr = read_ggcm_solarwind_file(
        ggcmpy.sample_dir / "cir07_19970227_liang_norcm/tmp.minvar/om1997.rr"
    )
    assert set(rr.columns) == {"om1997.rr"}
    assert pd.api.types.is_datetime64_ns_dtype(rr.index)
    assert np.isclose(rr["om1997.rr"].mean(), 4.8191107)


def test_read_ggcm_solarwind_directory():
    bfield = ggcmpy.timeseries.read_ggcm_solarwind_directory(
        ggcmpy.sample_dir / "cir07_19970227_liang_norcm/tmp.minvar", glob="om1997.b*gse"
    )
    assert set(bfield.columns) == {"om1997.bxgse", "om1997.bygse", "om1997.bzgse"}
    assert pd.api.types.is_datetime64_ns_dtype(bfield.index)
    assert np.isclose(bfield["om1997.bzgse"].mean(), -1.87348468)


def test_store_to_pyspedas():
    pyspedas = pytest.importorskip("pyspedas")

    bfield = ggcmpy.timeseries.read_ggcm_solarwind_directory(
        ggcmpy.sample_dir / "cir07_19970227_liang_norcm/tmp.minvar", glob="om1997.b*gse"
    )
    ggcmpy.timeseries.store_to_pyspedas(bfield)
    data = pyspedas.data_quants
    assert "om1997.bxgse" in data



from __future__ import annotations

import os
import pathlib
from collections.abc import Mapping
from typing import Any, Protocol

import numpy as np
from typing_extensions import Never, override
from xarray.backends import CachingFileManager, FileManager
from xarray.backends.common import AbstractDataStore
from xarray.backends.locks import SerializableLock, ensure_lock
from xarray.core import indexing
from xarray.core.utils import FrozenDict
from xarray.core.variable import Variable

from ggcmpy import openggcm

from .backends import jrrle
from .jrrle_array import JrrleArray

# not sure this is needed
JRRLE_LOCK = SerializableLock()

_OPENGGCM_UNITS = {
    # magnetosphere
    "rr": {
        "units": "1/cm^3",
        "long_name": "plasma number density",
    },
    "pp": {
        "units": "pPa",
        "long_name": "plasma pressure",
    },
    "vx": {
        "units": "km/s",
        "long_name": "plasma velocity, x-component",
    },
    "vy": {
        "units": "km/s",
        "long_name": "plasma velocity, y-component",
    },
    "vz": {
        "units": "km/s",
        "long_name": "plasma velocity, z-component",
    },
    "bx": {
        "units": "nT",
        "long_name": "magnetic field, x-component",
    },
    "by": {
        "units": "nT",
        "long_name": "magnetic field, y-component",
    },
    "bz": {
        "units": "nT",
        "long_name": "magnetic field, z-component",
    },
    "xjx": {
        "units": "uA/m^2",
        "long_name": "current density, x-component",
    },
    "xjy": {
        "units": "uA/m^2",
        "long_name": "current density, y-component",
    },
    "xjz": {
        "units": "uA/m^2",
        "long_name": "current density, z-component",
    },
    "resis": {
        "units": "Ohm m",
        "long_name": "plasma resistivity",
    },
    # ionosphere
    "pot": {
        "units": "V",
        "long_name": "Potential",
    },
    "pacurr": {
        "units": "uA/m^2",
        "long_name": "Field-Aligned Current Density (positive down)",
    },
    "sigh": {
        "units": "S",
        "long_name": "Hall Conductance",
    },
    "sigp": {
        "units": "S",
        "long_name": "Pedersen Conductance",
    },
    "prec_e_fe_1": {
        "units": "W/m^2",
        "long_name": "diffuse auroral electron precipitation energy flux",
    },
    "prec_e_fe_2": {
        "units": "W/m^2",
        "long_name": "discrete auroral electron precipitation energy flux",
    },
    "prec_e_e0_1": {
        "units": "eV",
        "long_name": "diffuse auroral electron precipitation mean energy",
    },
    "prec_e_e0_2": {
        "units": "eV",
        "long_name": "discrete auroral electron precipitation mean energy",
    },
    "delphi": {
        "units": "V",
        "long_name": "Knight potential",
    },
    "ppio": {
        "units": "pPa",
        "long_name": "mapped pressure",
    },
    "rrio": {
        "units": "1/cm^3",
        "long_name": "mapped density",
    },
    "ttio": {
        "units": "K",
        "long_name": "mapped temperature",
    },
    "fac_dyn": {
        "units": "uA/m^2",  # CHECKME
        "long_name": "dynamo field aligned current",
    },
    "fac_tot": {
        "units": "uA/m^2",  # CHECKME
        "long_name": "total field aligned current",
    },
    "xjh": {
        "units": "W/m^2",
        "long_name": "Joule heating rate",
    },
    "delbt": {
        "units": "T",
        "long_name": "ground magnetic perturbation",
    },
    "epio": {
        "units": "mV/m",
        "long_name": "azimuthal electric field",
    },
    "etio": {
        "units": "mV/m",
        "long_name": "meriodional electric field",
    },
}

_CRD_NAME = {
    "x": "-GSE_x",
    "y": "GSE_y",
    "z": "GSE_z",
}


class Lock(Protocol):
    """Provides duck typing for xarray locks, which do not inherit from a common base class."""

    def acquire(self, blocking: bool = True) -> bool: ...
    def release(self) -> None: ...
    def __enter__(self) -> None: ...
    def __exit__(self, *args: Any) -> None: ...
    def locked(self) -> bool: ...


class JrrleStore(AbstractDataStore):
    """DataStore to facilitate loading an OpenGGCM/jrrle2 file."""

    def __init__(
        self,
        manager: FileManager,
        filename: str | os.PathLike[Any],
        mode: str | None = None,
        lock: Lock = JRRLE_LOCK,
        autoclose: bool = False,
    ):
        assert isinstance(manager, FileManager)
        self._manager = manager
        self._mode = mode
        self.lock = ensure_lock(lock)  # type: ignore[no-untyped-call]
        self.autoclose = autoclose
        self._filename = filename

        self._meta = jrrle.parse_filename(self._filename)

    @classmethod
    def open(
        cls,
        filename: str | os.PathLike[Any],
        mode: str = "r",
        lock: Lock | None = None,
        autoclose: bool = False,
    ) -> JrrleStore:
        if lock is None:
            if mode == "r":
                lock = JRRLE_LOCK
            else:
                raise NotImplementedError()

        assert isinstance(filename, str | os.PathLike)

        manager = CachingFileManager(jrrle.JrrleFile, filename, mode=mode)
        return cls(manager, filename, mode=mode, lock=lock, autoclose=autoclose)

    def acquire(self, needs_lock: bool = True) -> jrrle.JrrleFile:
        with self._manager.acquire_context(needs_lock) as file:  # type: ignore[no-untyped-call]
            ds = file
        assert isinstance(ds, jrrle.JrrleFile)
        return ds

    @property
    def ds(self) -> jrrle.JrrleFile:
        return self.acquire()

    def dims(self) -> tuple[str, ...]:
        meta = self._meta
        if meta["type"] == "3df":
            return ("x", "y", "z")
        if meta["type"] == "2df":
            return tuple(dim for dim in ("x", "y", "z") if dim != meta["plane"])
        assert meta["type"] == "iof"
        return ("longs", "lats")

    def coords(self, shape: tuple[int, ...]) -> dict[str, Variable]:
        meta = self._meta
        if meta["type"] in {"2df", "3df"}:
            try:
                grid2_filename = pathlib.Path(meta["dirname"] / f"{meta['run']}.grid2")
                coords: dict[str, Any] = openggcm.read_grid2(grid2_filename)
            except FileNotFoundError:
                grid_filename = pathlib.Path(meta["dirname"] / f"{meta['run']}.grid")
                coords = openggcm.read_grid(grid_filename)
            if meta["type"] == "2df":
                coords[meta["plane"]] = [meta["plane_location"]]

            return {
                name: Variable(
                    dims=(name,),
                    data=data,
                    attrs={"units": "RE", "long_name": _CRD_NAME[name]},
                )
                for name, data in coords.items()
            }

        if meta["type"] == "iof":
            return {
                "lats": Variable(
                    dims=("lats",),
                    data=np.linspace(90.0, -90.0, shape[1]),
                    attrs={"units": "degrees_north", "long_name": "magnetic latitude"},
                ),
                "longs": Variable(
                    dims=("longs",),
                    data=np.linspace(-180.0, 180.0, shape[0]),
                    attrs={"units": "degrees_east", "long_name": "magnetic longitude"},
                ),
            }

        raise NotImplementedError

    def open_store_variable(
        self,
        name: str,
        fld_info: Mapping[str, Any],
    ) -> Variable:
        attrs = dict(fld_info)
        attrs.pop("time", None)

        data = indexing.LazilyIndexedArray(JrrleArray(name, self, fld_info))
        encoding: dict[str, Any] = {}

        # save source so __repr__ can detect if it's local or not
        encoding["source"] = self._filename
        encoding["original_shape"] = fld_info["shape"]
        encoding["dtype"] = np.dtype(np.float32)

        attrs.update(_OPENGGCM_UNITS.get(name, {}))

        return Variable(self.dims(), data, attrs, encoding)

    @override
    def get_variables(self) -> Mapping[str, Variable]:
        shape: tuple[int, ...] | None = None
        time: str | None = None
        inttime: int | None = None
        elapsed_time: float | None = None

        variables = dict[str, Variable]()
        for fld, fld_info in self.ds.vars.items():
            if shape is not None:
                assert shape == fld_info["shape"], "inconsistent shapes in jrrle file"
            if time is not None:
                assert time == fld_info["time"], "inconsistent time info in jrrle file"

            shape = fld_info["shape"]
            time = fld_info["time"]
            inttime = fld_info["inttime"]
            elapsed_time = fld_info["elapsed_time"]
            variables[fld] = self.open_store_variable(fld, fld_info)

        assert shape is not None
        variables.update(self.coords(shape))

        variables["time"] = Variable(("time"), [np.datetime64(time, "ns")])
        variables["inttime"] = Variable(("time"), [inttime])
        variables["elapsed_time"] = Variable(("time"), [elapsed_time])

        return FrozenDict(variables)

    @override
    def get_attrs(self) -> Mapping[str, Any]:
        return FrozenDict({"run": self._meta["run"]})

    @override
    def get_dimensions(self) -> Never:
        raise NotImplementedError()



from __future__ import annotations

import os
from collections.abc import Iterable
from typing import Any

from typing_extensions import override
from xarray.backends import BackendEntrypoint
from xarray.backends.common import AbstractDataStore
from xarray.backends.store import StoreBackendEntrypoint
from xarray.core.dataset import Dataset
from xarray.core.datatree import DataTree
from xarray.core.types import ReadBuffer

from .backends import jrrle
from .jrrle_store import JrrleStore


class JrrleEntrypoint(BackendEntrypoint):
    """Entrypoint that lets xarray recognize and read OpenGGCM jrrle (custom binary) output."""

    # url = "https://link_to/your_backend/documentation"  # FIXME

    open_dataset_parameters = ("filename_or_obj", "drop_variables")

    def guess_can_open(self, filename_or_obj):
        if not isinstance(filename_or_obj, str | os.PathLike):
            return False

        try:
            jrrle.parse_filename(filename_or_obj)
        except ValueError:
            return False

        return True

    @override
    def open_dataset(
        self,
        filename_or_obj: str | os.PathLike[Any] | ReadBuffer[Any] | AbstractDataStore,
        *,
        mask_and_scale: bool = True,
        decode_times: bool = True,
        concat_characters: bool = True,
        decode_coords: bool = True,
        drop_variables: str | Iterable[str] | None = None,
        use_cftime: bool | None = None,
        decode_timedelta: bool | None = None,
    ) -> Dataset:
        if isinstance(filename_or_obj, str | os.PathLike):
            store = JrrleStore.open(filename_or_obj)
        else:
            msg = f"unknown {filename_or_obj=}"
            raise TypeError(msg)

        store_entrypoint = StoreBackendEntrypoint()

        return store_entrypoint.open_dataset(
            store,
            mask_and_scale=mask_and_scale,
            decode_times=decode_times,
            concat_characters=concat_characters,
            decode_coords=decode_coords,
            drop_variables=drop_variables,
            use_cftime=use_cftime,
            decode_timedelta=decode_timedelta,
        )

    @override
    def open_datatree(
        self,
        filename_or_obj: str | os.PathLike[Any] | ReadBuffer[Any] | AbstractDataStore,
        **kwargs: Any,
    ) -> DataTree:
        raise NotImplementedError()



from __future__ import annotations

import os
from pathlib import Path
from typing import Any


def parse_filename(filename: str | os.PathLike[Any]) -> Any:
    filename = Path(filename)
    dirname = filename.parent
    run, ifx, step = filename.name.split(".")
    meta = {"dirname": dirname, "run": run, "step": int(step)}
    if ifx.startswith(("px_", "py_", "pz_")):
        meta["type"] = "2df"
        meta["plane"] = ifx[1]
        # given as integer in tenths of RE
        meta["plane_location"] = float(ifx[3:]) / 10
    else:
        meta["type"] = ifx
    return meta
